import 'package:wowit/Widget/filter_widget/filter_widget.dart';
import 'package:wowit/Widget/filter_widget/filter_widget_exception.dart';
import 'package:wowit/model.dart';

class FilterController {
  final Map<String, Object> mapCondition = {};
  final Map<String, ESortMode> mapSortation = {};
  final List<FilterItem> filterItems;

  FilterController(this.filterItems) {
    _addDefaultValues();
  }

  // mapCondition khi filter
  // Xử lý trường hợp rangeOrGreaterThanMax
  Map<String, Object> get filterMapCondition {
    final rangeOrGreatFitlerItems = filterItems
        .where((element) =>
            element.filterMode == EFilterMode.inRangeOrGreaterThanMax)
        .toList();

    if (rangeOrGreatFitlerItems.isNotEmpty) {
      final Map<String, Object> filterMapCondition = {};
      filterMapCondition.addEntries(mapCondition.entries);
      for (final filterItem in rangeOrGreatFitlerItems) {
        if (filterItem.maxValue == null) {
          throw FilterWidgetException(
              'Please provide maxValue for FilterItem with fieldName = ${filterItem.fieldName}');
        }
        final rangeValue = mapCondition[filterItem.fieldName] as RangeValue;
        if (rangeValue != null && rangeValue.end >= filterItem.maxValue) {
          filterMapCondition.update(
              filterItem.fieldName, (_) => RangeValue(rangeValue.start, null));
        }
      }
      return filterMapCondition;
    }
    return mapCondition;
  }

  void updateMapCondition(String fieldName, Object value) {
    mapCondition.update(
      fieldName,
      (_) => value,
      ifAbsent: () => value,
    );
  }

  void updateMapSortation(String fieldName, ESortMode sortMode) {
    mapSortation.clear();
    mapSortation.addAll({fieldName: sortMode});
  }

  List<Map<String, Object>> applySort(List<Map<String, Object>> data) {
    if (mapSortation.isEmpty) {
      return data;
    }

    final fieldName = mapSortation.keys.first;
    final sortMode = mapSortation[fieldName];

    List<Map<String, Object>> sortedData = List.from(data);

    sortedData.sort((a, b) {
      switch (sortMode) {
        case ESortMode.smallestToLargest:
          return _compareTo(a[fieldName], b[fieldName]);
        case ESortMode.largestToSmallest:
          return _compareTo(b[fieldName], a[fieldName]);
      }

      throw FilterWidgetException('SortMode is invalid');
    });

    filterItems.forEach((filterItem) {});

    return sortedData;
  }

  int _compareTo(dynamic value1, dynamic value2) {
    try {
      if (value2 == null) return 1;
      if (value1 == null) return -1;
      return value1.compareTo(value2);
    } catch (e) {
      throw FilterWidgetException(
          'the data used to sort does not have method compareTo');
    }
  }

  List<Map<String, Object>> applyFilter(List<Map<String, Object>> data) {
    if (data == null) return [];
    List<Map<String, Object>> filteredData = data;

    filterItems.forEach((filterItem) {
      filteredData = filteredData.where((dataItem) {
        if (!dataItem.containsKey(filterItem.fieldName)) {
          throw FilterWidgetException(
              'Filter Error: data not contain fieldname \'${filterItem.fieldName}\'');
        }

        if (!mapCondition.containsKey(filterItem.fieldName) ||
            mapCondition[filterItem.fieldName] == null ||
            mapCondition[filterItem.fieldName] == filterItem.allValue) {
          return true;
        }

        if (dataItem[filterItem.fieldName] == null) return false;

        switch (filterItem.filterMode) {
          case EFilterMode.contain:
            return _checkContain(filterItem, dataItem);
          case EFilterMode.equal:
            return _checkEqual(filterItem.fieldName, dataItem);
          case EFilterMode.lessThan:
            return _compared(filterItem, dataItem, (a, b) => a < b);
          case EFilterMode.greaterThan:
            return _compared(filterItem, dataItem, (a, b) => a > b);
          case EFilterMode.range:
            return _checkRange(filterItem, dataItem);
          case EFilterMode.lessOrEqual:
            return _compared(filterItem, dataItem, (a, b) => a <= b);
          case EFilterMode.greaterOrEqual:
            return _compared(filterItem, dataItem, (a, b) => a >= b);
          case EFilterMode.insensitiveAndUnsignedContain:
            return _checkInsensitiveAndUnsignedContain(filterItem, dataItem);
          case EFilterMode.inRangeOrGreaterThanMax:
            return _checkRangeOrGreaterThanMax(filterItem, dataItem);
            break;
        }

        return true;
      }).toList();
    });

    return applySort(filteredData);
  }

  void clear() {
    mapCondition.clear();
    _addDefaultValues();
  }

  void _addDefaultValues() {
    filterItems.forEach((item) {
      if (item.defaultValue != null) {
        mapCondition.addAll({item.fieldName: item.defaultValue});
      }
    });
  }

  bool _checkInsensitiveAndUnsignedContain(
    FilterItem filterItem,
    Map<String, Object> dataItem,
  ) {
    if (mapCondition[filterItem.fieldName] is String &&
        dataItem[filterItem.fieldName] is String) {
      final stringDataItem =
          getUnsignedVietnamese(dataItem[filterItem.fieldName] as String)
              .toLowerCase();
      final stringDataMapCondition =
          getUnsignedVietnamese(mapCondition[filterItem.fieldName] as String)
              .toLowerCase();

      return stringDataItem.contains(stringDataMapCondition);
    }

    throw FilterWidgetException(
        'fieldName \'${filterItem.fieldName}\' is not String');
  }

  bool _checkRange(FilterItem filterItem, Map<String, Object> data) {
    if (mapCondition[filterItem.fieldName] is RangeValue) {
      final rangeValues = mapCondition[filterItem.fieldName] as RangeValue;
      final value = data[filterItem.fieldName] as num;
      return value >= rangeValues.start && value <= rangeValues.end;
    }
    throw FilterWidgetException(
        'fieldName \'${filterItem.fieldName}\' is not RangeValue');
  }

  bool _checkRangeOrGreaterThanMax(
      FilterItem filterItem, Map<String, Object> data) {
    if (mapCondition[filterItem.fieldName] is RangeValue) {
      final rangeValues = mapCondition[filterItem.fieldName] as RangeValue;
      final value = data[filterItem.fieldName] as num;
      if (rangeValues.end >= filterItem.maxValue) {
        return value >= rangeValues.start;
      } else {
        return value >= rangeValues.start && value <= rangeValues.end;
      }
    }
    throw FilterWidgetException(
        'fieldName \'${filterItem.fieldName}\' is not RangeValue');
  }

  bool _checkEqual(String fieldName, Map<String, Object> data) {
    if (data[fieldName].runtimeType != mapCondition[fieldName].runtimeType) {
      print(
          '''FilterWidget warning: \'$fieldName\' data type is not same as mapCondition type 
          (${data[fieldName].runtimeType} # ${mapCondition[fieldName].runtimeType}''');
    }
    return data[fieldName] == mapCondition[fieldName];
  }

  bool _checkContain(FilterItem filterItem, Map<String, Object> data) {
    if (mapCondition[filterItem.fieldName] is String &&
        data[filterItem.fieldName] is String) {
      return (data[filterItem.fieldName] as String)
          .contains(mapCondition[filterItem.fieldName] as String);
    } else if (mapCondition[filterItem.fieldName] is List) {
      if ((mapCondition[filterItem.fieldName] as List).isEmpty) {
        return true;
      }

      return (mapCondition[filterItem.fieldName] as List)
          .contains(data[filterItem.fieldName]);
    }

    throw FilterWidgetException(
        'fieldName \'${filterItem.fieldName}\' is not String or List');
  }

  bool _compared(FilterItem filterItem, Map<String, Object> data,
      bool Function(num valueA, num valueB) compareTo) {
    if (data[filterItem.fieldName] == null) return false;

    if (mapCondition[filterItem.fieldName] is num &&
        data[filterItem.fieldName] is num) {
      return compareTo(
        (data[filterItem.fieldName] as num),
        (mapCondition[filterItem.fieldName] as num),
      );
    }

    throw FilterWidgetException(
        'fieldName \'${filterItem.fieldName}\' is not num');
  }

  String getUnsignedVietnamese(String text) {
    final _vietnamese = 'aAeEoOuUiIdDyY';
    final _vietnameseRegex = <RegExp>[
      RegExp(r'à|á|ạ|ả|ã|â|ầ|ấ|ậ|ẩ|ẫ|ă|ằ|ắ|ặ|ẳ|ẵ'),
      RegExp(r'À|Á|Ạ|Ả|Ã|Â|Ầ|Ấ|Ậ|Ẩ|Ẫ|Ă|Ằ|Ắ|Ặ|Ẳ|Ẵ'),
      RegExp(r'è|é|ẹ|ẻ|ẽ|ê|ề|ế|ệ|ể|ễ'),
      RegExp(r'È|É|Ẹ|Ẻ|Ẽ|Ê|Ề|Ế|Ệ|Ể|Ễ'),
      RegExp(r'ò|ó|ọ|ỏ|õ|ô|ồ|ố|ộ|ổ|ỗ|ơ|ờ|ớ|ợ|ở|ỡ'),
      RegExp(r'Ò|Ó|Ọ|Ỏ|Õ|Ô|Ồ|Ố|Ộ|Ổ|Ỗ|Ơ|Ờ|Ớ|Ợ|Ở|Ỡ'),
      RegExp(r'ù|ú|ụ|ủ|ũ|ư|ừ|ứ|ự|ử|ữ'),
      RegExp(r'Ù|Ú|Ụ|Ủ|Ũ|Ư|Ừ|Ứ|Ự|Ử|Ữ'),
      RegExp(r'ì|í|ị|ỉ|ĩ'),
      RegExp(r'Ì|Í|Ị|Ỉ|Ĩ'),
      RegExp(r'đ'),
      RegExp(r'Đ'),
      RegExp(r'ỳ|ý|ỵ|ỷ|ỹ'),
      RegExp(r'Ỳ|Ý|Ỵ|Ỷ|Ỹ')
    ];

    String result = text;
    for (var i = 0; i < _vietnamese.length; ++i) {
      result = result.replaceAll(_vietnameseRegex[i], _vietnamese[i]);
    }
    return result;
  }
}
